# KNN回归器
from sklearn.neighbors import KNeighborsRegressor
# KNN分类器
from sklearn.neighbors import KNeighborsClassifier

from itproject.machinelearning.KNNRegressor import KNNRegressor
from itproject.utils import DataUtil, ChartUtil, AnalyseUtil

knn_regressor = KNeighborsRegressor()
myknn_regressor = KNNRegressor()
knn_classifier = KNeighborsClassifier()
knnType = "R"


def fit(X_train, y_train, knn_type="R",self=False):
    """用训练数据训练模型,默认是用KNN回归器，type=C表示用KNN分类器, self=False表示用Numpy的库，self=true表示用自定义的"""
    global knn_regressor
    global myknn_regressor
    global knn_classifier
    global knnType
    knnType = knn_type
    if knn_type == "R":
        if not self:
            knn_regressor.fit(X_train, y_train)
        else:
            myknn_regressor.fit(X_train, y_train)
    else:
        knn_classifier.fit(X_train, y_train)


def predict(X_test):
    """预测"""
    print("预测用的算法：{}".format(knnType))
    if knnType == "R":
        return knn_regressor.predict(X_test)
    else:
        return knn_classifier.predict(X_test)



import numpy as np

from itproject.utils import ArrayUtil

"""
矩阵工具包，实现矩阵加减乘除等运算,切片思想，广播机制理解
"""


def getOnes(m, n):
    """获取m*n全是1的矩阵，exp：X = getOnes(3, 2)"""
    return np.ones(shape=(m, n))


def getZeros(m, n):
    return np.zeros(shape=(m, n))


def getEye(N, M=None):
    """获取N行的单位矩阵/对角矩阵, exp:X = getEye(3)"""
    return np.eye(N, M)


def getMNMatrix(m, n):
    """获取一个M*N的矩阵，m=-1表示只指定列数，n=-1表示只指定行数"""
    return np.arange(m * n).reshape(m, n)


def getImgMatrix(count, rgb=3, width=None, height=None):
    """获取count张宽是width高是height的图像的矩阵,exp: getImgMatrix(3,width=10,height=10)"""
    return np.random.randn(count, rgb, width, height)


def add(np_flag, X1, X2, *args, **kwargs):
    """
    两个矩阵相加,np_flag=true将调用numpy封装的方法
    """
    if np_flag:
        return np.add(X1, X2, *args, **kwargs)
    else:
        return A + B


def multi(A, B):
    """两个矩阵相乘"""
    return A @ B


def getMax(X, axis=None):
    """获取矩阵某行或某列的最大值, 0代表按列，1代表按行 exp: getMax(X, axis=0)"""
    return np.array(X).max(axis=axis)


def getMaxIdx(X, axis=None):
    """获取矩阵某行或某列的最大值对应的索引, 0代表按列，1代表按行 exp: getMaxIdx(X, axis=1)"""
    return np.array(X).argmax(axis=axis)


def getSubMatrix(X, row=None, col=None):
    """切片获取某一行或某一列,或者某个元素,如果row或col是负数，表示从后面开始算起
    exp:getSubMatrix(X,row=2), getSubMatrix(X, row=1,col=-1)行与列从1开始算"""
    if row is None and col is None:
        return X
    elif col is None:
        if row > 0:
            return np.array(X)[row - 1, :]
        else:
            return np.array(X)[row, :]
    elif row is None:
        if col > 0:
            return np.array(X)[:, col - 1]
        else:
            return np.array(X)[:, col]
    else:
        if row > 0 and col > 0:
            return np.array(X)[row - 1, col - 1]
        elif row < 0 and col < 0:
            return np.array(X)[row, col]
        elif row < 0 and col > 0:
            return np.array(X)[row, col - 1]
        else:
            return np.array(X)[row - 1, col]


def getMatrixT(arr):
    """获取转置矩阵,exp: getMatrixT(X)"""
    return np.array(arr).T



import numpy as np


"""
    自定义KNN回归器
"""


class KNNRegressor(object):
    """
        自定义KNN回归器
         ....
    """

    def __init__(self, n_neighbors=5):
        """
            初始化函数
            - 超参数
        """
        self.n_neighbors = n_neighbors

    def fit(self, X, y):
        """
            训练过程
            传入：训练数据
        """

        self.X = np.array(X)
        self.y = np.array(y)

        # 参数校验
        if self.X.ndim != 2 or self.y.ndim != 1:
            raise Exception("训练数据的维度有错 ..")

    def predict(self, X):
        """
            预测过程
            传入：待预测的特征
            注意：批量预测

            1，找K个好友；
            2，取均值
        """
        # 类型转换
        X = np.array(X)

        # 参数校验
        if X.ndim != 2 or self.X.shape[1] != X.shape[1]:
            raise Exception("默认是批量预测，特征维度必须是二维的 ...")

        # 定义结果列表
        result = []

        # 计算过程
        for x in X:
            # 计算待测样本 x 和整个数据集中样本 self.X 的距离
            distances = np.sqrt(((self.X - x) ** 2).sum(axis=1))

            np.sqrt(((self.X - x) ** 2).sum(axis=1))

            # 找到K个最近的邻居
            idxes = np.argsort(distances)[:self.n_neighbors]
            # 求K个最近邻居的均值
            result.append(self.y[idxes].mean())

        # 返回结果
        return np.array(result)



from matplotlib import pyplot as plt

"""数据可视化工具，可以用于预测结果定性分析"""


def plot(X, y):
    """绘制折线图"""
    plt.plot(X, y)
    plt.show()


def scatter(X, y):
    """绘制散点图"""
    plt.scatter(X, y)
    plt.show()



from sklearn.datasets import load_boston, load_iris, load_breast_cancer
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

"""数据集工具，获取波士顿房价，获取鸢尾花等数据集"""


def boston():
    """获取波士顿房价数据集exp: X,y = boston()"""
    data_url = "http://lib.stat.cmu.edu/datasets/boston"
    raw_df = pd.read_csv(data_url, sep="\\s+", skiprows=22, header=None)
    X = np.hstack([raw_df.values[::2, :], raw_df.values[1::2, :2]])
    y = raw_df.values[1::2, 2]
    return X, y


def iris():
    """获取鸢尾花数据集exp: X,y = iris()"""
    return load_iris(return_X_y=True)


def cancer():
    """获取癌症预测数据集, 用于分析是良性还是恶性exp: X, y = cancer()"""
    return load_breast_cancer(return_X_y=True)


def split(X, y, test_percentage=0.20):
    """切分数据集,默认80%用于训练，20%用于测试,exp: split(X, y)"""
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_percentage)
    return X_train, X_test, y_train, y_test



import numpy as np

"""一维数组/向量的运算，加减乘除，平均值，标准差，方差等等"""


def getArray(n):
    """获取n个数字的一维数组"""
    return np.arange(n)


def getArrayByZone(start=None, *args, **kwargs):
    """获取指定区间的数组，并且指定每个数之间的间隔, step为负数表示反方向生成
    exp: getArrayByZone(start=1,stop=12,step=2)"""
    return np.arange(start=start, *args, **kwargs)


def getArrayByCount(start, stop, num=50, endpoint=True, retstep=False, dtype=None,
                    axis=0):
    """指定某个区间，指定要生成多少个数，产生等间隔的数列
    exp: arr1 = getArrayByCount(start=1, stop=12, num=12)
    """
    return np.linspace(start=start, stop=stop, num=num, endpoint=True, retstep=False, dtype=None,
                       axis=0)


def getArrayAvg(arr):
    """求数组的平均值"""
    return np.array(arr).mean()


def getArrayStd(arr):
    """求数组的标准差"""
    return np.array(arr).std()


def getArrayVar(arr):
    """求数组的方差"""
    return np.array(arr).var()


def getArraySize(arr):
    """求数组的长度"""
    return np.array(arr).size


def getArrayShape(arr):
    """求数组的形状"""
    return np.array(arr).shape


def getArrayNDim(arr):
    """求数组的维度"""
    return np.array(arr).ndim


def getArrayMod(arr):
    """求向量的模，模就是向量的长度, 模=平方和开根号"""
    return np.sqrt((np.array(arr) ** 2).sum())


def getArrayDot(arr1, arr2):
    """求两个向量的点乘积, 即两个向量的内积"""
    return np.array(arr1).dot(np.array(arr2))


def sort(arr, axis=-1, kind=None, order=None):
    """数组排序："""
    return np.array(arr).argsort(axis=axis, kind=kind, order=order)


def distance(X1, X2):
    """计算两个向量之间的距离，距离越近表示这两个向量越相似, 初中知识是两点之间的距离
    高维空间里称为欧几里得距离，也叫欧氏距离
    exp:
    X = np.array([1,2,3])
    X1 = np.array([4,5,6])
    ArrayUtil.distance(X, X1)
    """
    return np.sqrt(((X1 - X2) ** 2).sum())


def cos(X1, X2):
    """计算两个向量的余弦相似度，两个向量的夹角越小，表示越相似"""
    return X1 @ X2 / np.linalg.norm(X1) / np.linalg.norm(X2)


def pierxun(X1, X2):
    """皮尔逊相关系数，本质也是余弦相似度，也是用来衡量两个向量的相似度"""
    pass



import numpy as np

"""预测结果定性分析，MAE,MSE等等"""


def mae(y_pred, y_test):
    """预测结果y_pred和实际数据y_test的平均绝对误差MAE, 用这种方式衡量，求导数是麻烦事"""
    return np.abs(y_pred - y_test).mean()


def mse(y_pred, y_test):
    """预测结果y_pred和实际数据y_test的均方差，解决MAE求导数麻烦的问题
    回归问题一般都是用MSE来衡量误差的
    """
    return ((y_pred - y_test) ** 2).mean()
